"""Common information so that all traceback generating scripts
   create files in the same format.

   IMPORTANT: this assumes that all tests files with names of the form

       raise_something.dialect

"""
import sys
from contextlib import redirect_stderr


def write(text):
    sys.stderr.write(text + "\n")


def make_title(text):
    write("\n" + text)
    write("-" * len(text) + "\n")
    write(".. code-block:: none\n")


all_imports = {
    "IfNobreakError": "raise_if_nobreak",
    "MismatchedBracketsError": "raise_mismatched_brackets",
    "MissingLeftBracketError": "raise_missing_left_bracket",
    "MissingRepeatColonError": "raise_missing_repeat_colon",
    "MissingRepeatError": "raise_missing_repeat",
    "NameError": "raise_name_error",
    "NobreakFirstError": "raise_nobreak_first",
    "NobreakSyntaxError": "raise_nobreak_syntax",
    "RepeatFirstError": "raise_repeat_first",
    "TryNobreakError": "raise_try_nobreak",
    "UnknownDialectError": "raise_unknown_dialect",
    "UnknownLanguageError": "raise_unknown_language",
}


def create_tracebacks(target, intro_text):
    with open(target, "w", encoding="utf8") as out:
        with redirect_stderr(out):
            write(intro_text)

            for title in all_imports:
                make_title(title)
                __import__(all_imports[title])



"""Creates a version of traceback_xx.rst to insert in the documentation.
"""

# When creating a new dialect, you need to:
# 1. Make a copy of this file
# 2. Change the value of LANG as well as 'content' so that they reflect the
#    appropriate language
# 3. Make a version of all the raise_some_exception tests in that dialect;
#    this is more easily done using a custom script which you should find
#    in the documentation

import os
import sys
import platform

LANG = "upper"

# Make it possible to find docs and tests source
this_dir = os.path.dirname(__file__)
root_dir = os.path.abspath(os.path.join(this_dir, ".."))
test_dir = os.path.abspath(os.path.join(this_dir, f"py{LANG}"))

sys.path.insert(0, root_dir)

# set up import hook
import avantpy  # noqa

sys.path.insert(0, this_dir)
sys.path.insert(0, test_dir)
os.chdir(test_dir)

import tb_common  # noqa

target = os.path.normpath(os.path.join(root_dir, f"docs/source/tracebacks_{LANG}.rst"))

intro_text = """Friendly error messages
=======================

AvantPy aims to provide friendlier feedback when an exception
is raised than what is done by Python.
Such feedback will also be available in languages other than English.

.. note::

     The content of this page is generated by running
     {name} located in the ``tests/`` directory.
     This needs to be done explicitly, independently of updating the
     documentation using Sphinx.

AvantPy version: {avantpy}
Python version: {python}

""".format(
    avantpy=avantpy.version.__version__, python=platform.python_version(), name=__file__
)


tb_common.create_tracebacks(target, intro_text)



import os
import sys

parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
if parent_dir not in sys.path:
    sys.path.insert(0, parent_dir)

# set up import hook
import avantpy  # noqa



def test_mismatchedbrackets(capsys):
    from . import raise_mismatched_brackets

    if not hasattr(raise_mismatched_brackets, "err"):
        info = str(capsys.readouterr())
        raise_mismatched_brackets.err = info
    assert "MismatchedBracketsError" in raise_mismatched_brackets.err



# content of conftest.py
import pytest


def pytest_collect_file(parent, path):
    if path.ext.startswith(".py") and (
        path.basename.startswith("test_") or path.basename.startswith("catch_")
    ):
        return pytest.Module(path, parent)



import subprocess

sessions = [
    (
        "python -m avantpy",
        "a = 41 \na+=1\nprint(a)\nd=e\n",
        ["42"],
        ["-->1: d=e", "NameError"],
    )
]


def test_console():
    """Function discoverable and run by pytest"""
    for command, inp, out, err in sessions:
        process = subprocess.Popen(
            command,
            shell=False,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,  # use strings as input
        )
        # I have found comparisons with stderr problematic, so I ignore it
        # However, since all tests are supposed not to raise exceptions
        # but only valid output, this does not impact the reliability
        # of these tests: if stdout is not as expected, we have a problem.
        stdout, stderr = process.communicate(inp)
        process.wait()
        for item in out:
            assert item in stdout
        for item in err:
            assert item in stderr


if __name__ == "__main__":
    test_console()



import os
from avantpy import converter


def test_transcoding():
    cwd = os.path.dirname(__file__)
    source_path = os.path.normpath(os.path.abspath(os.path.join(cwd, "pyfr/pyfr.pyfr")))
    assert os.path.isfile(source_path)

    with open(source_path, encoding="utf8") as f:
        original = f.read()

    english = converter.transcode(original, "pyfr", "pyen")
    french = converter.transcode(english, "pyen", "pyfr")

    assert french == original



"""Spanish dialect"""

pyes = {
    "False": "Falso",
    "None": "Nada",
    "True": "Verdadero",
    "and": "y",
    "as": "como",
    "assert": "afirmar",
    "async": "async",  # do not translate
    "await": "await",  # as these are not for beginners
    "break": "interrumpir",
    "class": "clase",
    "continue": "continuar",
    "def": "def",
    "del": "eliminar",
    "elif": "osi",
    "else": ["sino", "nointerrumpir"],
    "except": "excepto",
    "finally": "finalmente",
    "for": "para",
    "from": "de",
    "global": "global",
    "if": "si",
    "import": "importar",
    "in": "en",
    "is": "es",
    "lambda": "función",
    "nonlocal": "nolocal",
    "not": "no",
    "or": "o",
    "pass": "seguir",
    "raise": "elevar",
    "return": "retornar",
    "try": "intentar",
    "while": "mientras",
    "with": "con",
    "yield": "ceder",
    # Special loop keywords
    "repeat": "repetir",
    "forever": "parasiempre",
    "until": "hasta",
    # a few builtins useful for beginners
    "input": "entrada",
    "print": "imprimir",
    "range": "intervalo",
    "exit": "salir",  # useful for console - translation should be checked
    # a well-known python expression
    '__name__ == "__main__"': "noimportado",
}



"""Custom exceptions"""

from .my_gettext import gettext_lang


class AvantPyException(Exception):
    """Base class for custom exceptions."""

    def __init__(self, msg, args):
        _ = gettext_lang.lang
        self.msg = msg
        self.args = args
        self.params = args[0]
        self.friendly = {"header": _("AvantPy exception:")}
        self.add_cause()

    def __str__(self):
        return f"{self.__class__.__name__}: {self.msg}"

    def add_cause(self):
        self.friendly["generic"] = None


class AvantPySyntaxError(AvantPyException, SyntaxError):
    """Base class for custom exceptions that are SyntaxErrors"""

    def __init__(self, msg, args):
        super().__init__(msg, args)
        _ = gettext_lang.lang
        self.friendly["header"] = _("AvantPy syntax error:")
        self.filename = self.params["filename"]
        self.lineno = self.params["linenumber"]
        if "offset" in self.params:
            self.offset = self.params["offset"]
        else:
            self.offset = 1


class IfNobreakError(AvantPySyntaxError):
    """Raised if ``nobreak`` is used instead of ``else`` in an ``if`` statement"""

    def add_cause(self):
        _ = gettext_lang.lang
        self.friendly["generic"] = _(
            "The AvantPy '{nobreak_kwd}' keyword cannot be used in\n"
            "an '{if_kwd}/{elif_kwd}/{else_kwd}' clause (Python: if/elif/else).\n"
        ).format(
            nobreak_kwd=self.params["nobreak keyword"],
            if_kwd=self.params["if_kwd"],
            elif_kwd=self.params["elif_kwd"],
            else_kwd=self.params["else_kwd"],
        )


class MismatchedBracketsError(AvantPySyntaxError):
    """Raised if a left parenthesis '(', or square bracket '['
       or curly bracket '{', is closed by a right 'bracket' of a different type.

       In standard Python, this would be indicated as::

           SyntaxError: invalid syntax
    """

    def add_cause(self):
        _ = gettext_lang.lang
        self.friendly["generic"] = _(
            "The opening {open_bracket} does not match the closing {close_bracket}.\n"
        ).format(
            open_bracket=self.params["open_bracket"],
            close_bracket=self.params["close_bracket"],
        )


class MissingLeftBracketError(AvantPySyntaxError):
    """Raised if a right parenthesis ')', or square bracket ']'
       or curly bracket '}', is found without having a matching left bracket
       found previously.

       In standard Python, this would be indicated as::

           SyntaxError: invalid syntax
    """

    def add_cause(self):
        _ = gettext_lang.lang
        self.friendly["generic"] = _(
            "The closing {bracket} does not match anything.\n"
        ).format(bracket=self.params["bracket"])


class MissingRepeatColonError(AvantPySyntaxError):
    """Raised if a line beginning with ``repeat`` does not end with a colon.
    """

    def add_cause(self):
        _ = gettext_lang.lang
        self.friendly["generic"] = _(
            "A statement beginning with the '{repeat_kwd}' keyword must be on\n"
            "a single line ending with a colon (:) that indicates the beginning of\n"
            "an indented block of code, with no other colon appearing on that line.\n"
        ).format(repeat_kwd=self.params["repeat_kwd"])


class MissingRepeatError(AvantPySyntaxError):
    """Raised if ``until`` or ``forever`` is used without being preceeded by
       ``repeat``.
    """

    def add_cause(self):
        _ = gettext_lang.lang
        self.friendly["generic"] = _(
            "The AvantPy '{keyword}'' keyword can be used only when"
            "preceded by '{repeat_kwd}'.\n"
        ).format(keyword=self.params["keyword"], repeat_kwd=self.params["repeat_kwd"])


class NobreakFirstError(AvantPySyntaxError):
    """Raised if ``nobreak`` is used somewhere other than at the beginning of a line."""

    def add_cause(self):
        _ = gettext_lang.lang
        self.friendly["generic"] = _(
            "The AvantPy '{nobreak_kwd}' keyword can be used instead of '{else_kwd}'\n"
            "(Python: else) only when it begins a new statement in\n"
            "'{for_kwd}/{while_kwd}' loops (Python: for/while).\n\n"
        ).format(
            nobreak_kwd=self.params["nobreak keyword"],
            for_kwd=self.params["for_kwd"],
            while_kwd=self.params["while_kwd"],
            else_kwd=self.params["else_kwd"],
        )


class NobreakSyntaxError(AvantPySyntaxError):
    """Raised if ``nobreak`` is without a matching block."""

    def add_cause(self):
        _ = gettext_lang.lang
        self.friendly["generic"] = _(
            "The AvantPy '{nobreak_kwd}' keyword can only be used as a replacement\n"
            "of '{else_kwd}' (Python: else) with a matching '{for_kwd}' or\n"
            "'{while_kwd}' loop (Python: for/while).\n"
        ).format(
            nobreak_kwd=self.params["nobreak keyword"],
            for_kwd=self.params["for_kwd"],
            while_kwd=self.params["while_kwd"],
            else_kwd=self.params["else_kwd"],
        )


class RepeatFirstError(AvantPySyntaxError):
    """Raised if ``repeat`` is used somewhere other than at the beginning of a line."""

    def add_cause(self):
        _ = gettext_lang.lang
        self.friendly["generic"] = _(
            "The AvantPy '{repeat_kwd}' keyword can only be used to begin\n"
            "a new loop (Python: equivalent to 'for' or 'while' loop).\n"
        ).format(repeat_kwd=self.params["repeat keyword"])


class TryNobreakError(AvantPySyntaxError):
    """Raised if ``nobreak`` is used instead of ``else`` in ``try/except`` statement"""

    def add_cause(self):
        _ = gettext_lang.lang
        self.friendly["generic"] = _(
            "The AvantPy '{nobreak_kwd}' keyword cannot be used in\n"
            "a '{try_kwd}/{except_kwd}/{else_kwd}/{finally_kwd}' clause\n"
            "(Python: try/except/else/finally).\n"
        ).format(
            nobreak_kwd=self.params["nobreak keyword"],
            try_kwd=self.params["try_kwd"],
            except_kwd=self.params["except_kwd"],
            else_kwd=self.params["else_kwd"],
            finally_kwd=self.params["finally_kwd"],
        )


class UnknownDialectError(AvantPyException):
    """Raised when attempting to set ``dialect`` to unsupported value."""

    def add_cause(self):
        _ = gettext_lang.lang
        dialect = self.args[0]
        all_dialects = self.args[1]

        self.friendly["generic"] = _(
            "The following unknown dialect was requested: {dialect}.\n\n"
            "The known dialects are: {all_dialects}.\n"
        ).format(dialect=dialect, all_dialects=all_dialects)


class UnknownLanguageError(AvantPyException):
    """Raised when attempting to set ``lang`` to unsupported value."""

    def add_cause(self):
        _ = gettext_lang.lang
        lang = self.args[0]
        all_langs = self.args[1]

        self.friendly["generic"] = _(
            "The following unknown languages was requested: {lang}.\n\n"
            "The known dialects are: {all_langs}.\n"
        ).format(lang=lang, all_langs=all_langs)



def test_name_error(capsys):
    from . import raise_name_error

    if not hasattr(raise_name_error, "err"):
        info = str(capsys.readouterr())
        raise_name_error.err = info
    assert "NameError" in raise_name_error.err



"""The goal of exception_handling.py is to display information about
exceptions being raised in a way that is much easier to understand
for beginners than normal Python tracebacks.  These "friendly" tracebacks
are written so that they can easily be translated into any human language.
"""
import sys
import traceback

from .session import state
from .my_gettext import gettext_lang

ENABLED = True

# Many strings are common to the information provided by  exceptions handlers below.
# To avoid code duplication, we define them here.
# Furthermore, note that two versions are provided: one where we assume that
# the code is standard Python, and that no dialect-based keywords are used,
# and a second version specific to dialects.


def avantpy_exception_message():
    _ = gettext_lang.lang
    return _(
        "    AvantPy exception: {exception_name}\n\n"
        "    Error found in file '{filename}' on line {linenumber}.\n\n"
        "{partial_source}\n\n"
    )


def python_exception_message():
    _ = gettext_lang.lang
    return _(
        "    Python exception: \n"
        "        {python_display}\n\n"
        "    Error found in file '{filename}' on line {linenumber}.\n\n"
        "{partial_source}\n\n"
    )


def disable():
    """Disables custom exception handling"""
    global ENABLED
    ENABLED = False


def get_partial_source(source, begin, end, marks=None, nb_digits=None):
    """Extracts a few relevant lines from a source file.

       If the number of lines would exceed a certain limit (currently 7)
       the function is called recursively to only show the a few lines
       at the beginning and at the end.
    """
    if nb_digits is None:
        nb_digits = len(str(end))
    no_mark = "       {:%d}: " % nb_digits
    with_mark = "    -->{:%d}: " % nb_digits
    continuation = "           ..."

    if marks is None:
        marks = []

    result = []
    if end - begin > 7:
        # Suppose the source spans line numbers 1 to 12. Splitting it to show
        # partial result would show lines 1 to 3, followed by 10 to 12.
        # If we want the indentation of the first part to match the indentation
        # of the second part, we must specify the length of the
        # indentation of the first part.
        # See the traceback for MismatchedBracketsError in the
        # documentation for an example.
        result.append(
            get_partial_source(
                source, begin, begin + 2, marks=marks, nb_digits=nb_digits
            )
        )
        result.append(continuation)
        result.append(get_partial_source(source, end - 2, end, marks=marks))
    else:
        lines = source.split("\n")
        for index, line in enumerate(lines, start=1):
            if index < begin:
                continue
            if index > end:
                break
            if index in marks:
                result.append(with_mark.format(index) + line)
            else:
                result.append(no_mark.format(index) + line)

    return "\n".join(result)


def extract_filename_linenumber(exc):
    """Extracting filename and linenumber where an error occurred based
       on information from traceback."""

    last_tb_line = traceback.format_tb(exc.__traceback__)[-1]

    # For last_tb_line, we expect something like
    # File "filename", line 1, in <module>
    splitted_line = last_tb_line.split(",")
    filename = splitted_line[0].replace("File ", "").replace('"', "").strip()
    if filename in ["<string>", "<tokenize>"] and state.current_filename is not None:
        filename = state.current_filename
    linenumber = int(splitted_line[1].replace("line ", ""))
    return filename, linenumber


def write_err(msg):
    """Writes a string to sys.stderr."""
    sys.stderr.write(msg)


def write_exception_info(exc, source):
    """Writes the information we have after processing the exception."""
    _ = gettext_lang.lang
    msg = handle_exception(exc, source)
    if msg is not None:
        write_err(msg)
    else:
        write_err(
            _("An exception was raised for which we have no simplified traceback:\n")
        )
        write_err("name: %s\n" % exc.__class__.__name__)
        write_err("args: " + str(exc.args) + "\n\n")
        write_err("Python traceback follows:\n\n")
        for item in traceback.format_tb(exc.__traceback__):
            write_err(item)
        write_err("\n")


def handle_exception(exc, source):
    """Generic function to handle exceptions and returns a
       friendlier traceback than Python.

       If the exception is not recognized as one for which a friendlier
       traceback can be provided, None is returned
    """
    if not ENABLED:
        # Let normal Python traceback through
        raise exc

    name = exc.__class__.__name__
    if name in dispatch:
        return dispatch[name](exc, source)
    else:
        return None  # just want to be explicit


# ================= Specific handlers below


def handle_IfNobreakError(exc, source):
    """Handles situation where ``nobreak`` was wrongly use as a
       replacement for ``else`` in an if/elif/else block.
    """
    _ = gettext_lang.lang
    params = exc.args[0]
    begin = params["if_linenumber"]
    end = params["linenumber"]
    marks = [end]
    partial_source = get_partial_source(source, begin, end, marks=marks)
    dialect = params["dialect"]

    message = avantpy_exception_message() + _(
        "    The AvantPy '{nobreak_kwd}' keyword cannot be used in\n"
        "    an '{if_kwd}/{elif_kwd}/{else_kwd}' clause (Python: if/elif/else).\n\n"
    )

    return message.format(
        exception_name=exc.__class__.__name__,
        filename=params["filename"],
        nobreak_kwd=params["nobreak keyword"],
        partial_source=partial_source,
        linenumber=end,
        dialect=dialect,
        if_kwd=params["if_kwd"],
        elif_kwd=params["elif_kwd"],
        else_kwd=params["else_kwd"],
    )


def handle_IndentationError(exc, source):
    """Handles IndentationError.
    """
    _ = gettext_lang.lang
    msg = exc.args[0]
    if "unexpected indent" in msg:
        this_case = _(
            "    In this case, the line indicated by an arrow\n"
            "    is more indented than expected and does not match\n"
            "    the indentation of the previous line.\n"
        )
    elif "expected an indented block" in msg:
        this_case = _(
            "    In this case, the line indicated by an arrow\n"
            "    was expected to begin a new indented block.\n"
        )
    else:
        this_case = _(
            "    In this case, the line indicated by an arrow\n"
            "    is less indented the preceding one,\n"
            "    and is not aligned vertically with another block of code.\n"
        )

    exc_name = exc.__class__.__name__
    python_display = "{exc_name}: {msg}".format(exc_name=exc_name, msg=msg)
    filename, linenumber, _ignore, _ignore = exc.args[1]

    if filename in ["<string>", "<tokenize>"] and state.current_filename is not None:
        filename = state.current_filename

    begin = linenumber - 4
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    dialect = state.current_dialect

    message_end = _(
        "    An indentation error occurs when a given line is\n"
        "    not indented (aligned vertically) as expected.\n"
        "{this_case}\n"
    )

    message = python_exception_message() + message_end

    return message.format(
        exception_name=exc_name,
        filename=filename,
        python_display=python_display,
        partial_source=partial_source,
        dialect=dialect,
        linenumber=linenumber,
        this_case=this_case,
    )


def handle_MismatchedBracketsError(exc, source):
    """Handles situation where a bracket of one kind '([{'
       is closed with a bracket of a different kind.
    """
    _ = gettext_lang.lang
    params = exc.args[0]
    begin = params["open_linenumber"]
    end = params["linenumber"]
    marks = [begin, end]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    dialect = params["dialect"]

    message_end = _(
        "    The opening {open_bracket} does not match the closing {close_bracket}.\n\n"
    )

    message = avantpy_exception_message() + message_end

    return message.format(
        exception_name=exc.__class__.__name__,
        filename=params["filename"],
        linenumber=end,
        open_bracket=params["open_bracket"],
        close_bracket=params["close_bracket"],
        partial_source=partial_source,
        dialect=dialect,
    )


def handle_MissingLeftBracketError(exc, source):
    """Handles situation where at bracket of one kind ),],}
       is closed without a corresponding open bracket.
    """
    _ = gettext_lang.lang
    params = exc.args[0]
    linenumber = params["linenumber"]
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    dialect = params["dialect"]

    message_end = _("    The closing {bracket} does not match anything.\n\n")

    message = avantpy_exception_message() + message_end

    return message.format(
        exception_name=exc.__class__.__name__,
        filename=params["filename"],
        partial_source=partial_source,
        dialect=dialect,
        linenumber=linenumber,
        bracket=params["bracket"],
    )


def handle_MissingRepeatColonError(exc, source):
    """Handles situation where a statement beginning with repeat
       does not end with a colon.
    """
    _ = gettext_lang.lang
    params = exc.args[0]
    linenumber = params["linenumber"]
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    message_end = _(
        "    A statement beginning with the '{repeat_kwd}' keyword must be on\n"
        "    a single line ending with a colon (:) that indicates the beginning of\n"
        "    an indented block of code, with no other colon appearing on that line.\n\n"
    )
    message = avantpy_exception_message() + message_end

    return message.format(
        exception_name=exc.__class__.__name__,
        filename=params["filename"],
        repeat_kwd=params["repeat_kwd"],
        partial_source=partial_source,
        linenumber=linenumber,
        dialect=params["dialect"],
    )


def handle_MissingRepeatError(exc, source):
    """Handles situation where either ``until`` or ``forever`` was
       used without being preceeded by ``repeat``.
    """
    _ = gettext_lang.lang
    params = exc.args[0]
    linenumber = params["linenumber"]
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    message_end = _(
        "    The AvantPy '{keyword}'' keyword can be used only when"
        " preceded by '{repeat_kwd}'.\n\n"
    )

    message = avantpy_exception_message() + message_end
    return message.format(
        exception_name=exc.__class__.__name__,
        filename=params["filename"],
        keyword=params["keyword"],
        partial_source=partial_source,
        linenumber=linenumber,
        dialect=params["dialect"],
        repeat_kwd=params["repeat_kwd"],
    )


def handle_NameError(exc, source):
    """Handles situation where a NameError is raise."""
    _ = gettext_lang.lang
    msg = exc.args[0]
    var_name = msg.split("'")[1]
    exc_name = exc.__class__.__name__
    python_display = "{exc_name}: {msg}".format(exc_name=exc_name, msg=msg)

    filename, linenumber = extract_filename_linenumber(exc)

    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    message_end = _(
        "    A NameError exception indicates that a variable or\n"
        "    function name is not known to Python.\n"
        "    Most often, this is because there is a spelling mistake; however,\n"
        "    sometimes it is because it is used before being defined\n"
        "    or given a value. In your program, the unknown name is '{var_name}'.\n\n"
    )
    message = python_exception_message() + message_end

    return message.format(
        exception_name=exc.__class__.__name__,
        filename=filename,
        python_display=python_display,
        var_name=var_name,
        partial_source=partial_source,
        linenumber=linenumber,
        dialect=state.current_dialect,
    )


def handle_NobreakFirstError(exc, source):
    """Handles situation where ``nobreak`` was wrongly use as a
       replacement for ``else`` in statements like
       var = x if condition else y.
    """
    _ = gettext_lang.lang
    params = exc.args[0]
    linenumber = params["linenumber"]
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    message_end = _(
        "    The AvantPy '{nobreak_kwd}' keyword can be used instead of '{else_kwd}'\n"
        "    (Python: else) only when it begins a new statement in\n"
        "    '{for_kwd}/{while_kwd}' loops (Python: for/while).\n\n"
    )
    message = avantpy_exception_message() + message_end

    return message.format(
        exception_name=exc.__class__.__name__,
        filename=params["filename"],
        nobreak_kwd=params["nobreak keyword"],
        partial_source=partial_source,
        linenumber=linenumber,
        for_kwd=params["for_kwd"],
        while_kwd=params["while_kwd"],
        else_kwd=params["else_kwd"],
    )


def handle_NobreakSyntaxError(exc, source):
    """Handles situation where ``nobreak`` is used without a matching
       ``for`` or ``while`` loop and not already raised.
    """
    _ = gettext_lang.lang
    params = exc.args[0]
    linenumber = params["linenumber"]
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    message_end = _(
        "    The AvantPy '{nobreak_kwd}' keyword can only be used as a replacement\n"
        "    of '{else_kwd}' (Python: else) with a matching '{for_kwd}' or\n"
        "    '{while_kwd}' loop (Python: for/while).\n\n"
    )
    message = avantpy_exception_message() + message_end

    return message.format(
        exception_name=exc.__class__.__name__,
        filename=params["filename"],
        nobreak_kwd=params["nobreak keyword"],
        partial_source=partial_source,
        linenumber=linenumber,
        for_kwd=params["for_kwd"],
        while_kwd=params["while_kwd"],
        else_kwd=params["else_kwd"],
    )


def handle_RepeatFirstError(exc, source):
    """Handles situation where ``repeat`` was uses wrongly as
       it did not appear as the first keyword of a given statement.
    """
    _ = gettext_lang.lang
    params = exc.args[0]
    linenumber = params["linenumber"]
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    message_end = _(
        "    The AvantPy '{repeat_kwd}' keyword can only be used to begin\n"
        "    a new loop (Python: equivalent to 'for' or 'while' loop).\n\n"
    )
    message = avantpy_exception_message() + message_end

    return message.format(
        exception_name=exc.__class__.__name__,
        filename=params["filename"],
        repeat_kwd=params["repeat keyword"],
        partial_source=partial_source,
        linenumber=linenumber,
    )


def handle_TabError(exc, source):
    """Handles TabError.
    """
    _ = gettext_lang.lang
    msg = exc.args[0]

    exc_name = exc.__class__.__name__
    python_display = "{exc_name}: {msg}".format(exc_name=exc_name, msg=msg)
    filename, linenumber, _ignore, _ignore = exc.args[1]

    if filename in ["<string>", "<tokenize>"] and state.current_filename is not None:
        filename = state.current_filename

    begin = linenumber - 4
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    message_end = _(
        "    A TabError indicates that you have used both spaces\n"
        "    and tab characters to indent your code.\n"
        "    This is not allowed in Python.\n"
        "    Python's recommendation is to always use spaces to indent your code.\n\n"
    )

    message = python_exception_message() + message_end

    return message.format(
        exception_name=exc_name,
        filename=filename,
        python_display=python_display,
        partial_source=partial_source,
        linenumber=linenumber,
    )


def handle_TryNobreakError(exc, source):
    """Handles situation where ``nobreak`` was wrongly use as a
       replacement for ``else`` in an try/except/else/finally block.
    """
    _ = gettext_lang.lang
    params = exc.args[0]
    begin = params["try_linenumber"]
    end = params["linenumber"]
    marks = [end]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    message_end = _(
        "    The AvantPy '{nobreak_kwd}' keyword cannot be used in\n"
        "    a '{try_kwd}/{except_kwd}/{else_kwd}/{finally_kwd}' clause\n"
        "    (Python: try/except/else/finally).\n\n"
    )
    message = avantpy_exception_message() + message_end

    return message.format(
        exception_name=exc.__class__.__name__,
        filename=params["filename"],
        nobreak_kwd=params["nobreak keyword"],
        partial_source=partial_source,
        linenumber=end,
        try_kwd=params["try_kwd"],
        except_kwd=params["except_kwd"],
        else_kwd=params["else_kwd"],
        finally_kwd=params["finally_kwd"],
    )


def handle_UnknownDialectError(exc, *args):
    """Handles error raised when an unknown dialect is requested
    """
    _ = gettext_lang.lang
    dialect = exc.args[0]
    all_dialects = exc.args[1]

    message = _(
        "    AvantPy exception: UnknownDialectError\n\n"
        "    The following unknown dialect was requested: {dialect}.\n\n"
        "    The known dialects are: {all_dialects}.\n\n"
    )
    return message.format(
        exception_name=exc.__class__.__name__,
        dialect=dialect,
        all_dialects=all_dialects,
    )


def handle_UnknownLanguageError(exc, *args):
    """Handles error raised when an unknown language is requested
    """
    _ = gettext_lang.lang
    lang = exc.args[0]
    all_langs = exc.args[1]

    message = _(
        "    AvantPy exception: UnknownLanguageError\n\n"
        "    The following unknown language was requested: {lang}.\n\n"
        "    The known languages are: {all_langs}.\n\n"
    )

    return message.format(
        exception_name=exc.__class__.__name__, lang=lang, all_langs=all_langs
    )


dispatch = {
    "IfNobreakError": handle_IfNobreakError,
    "IndentationError": handle_IndentationError,
    "NameError": handle_NameError,
    "MismatchedBracketsError": handle_MismatchedBracketsError,
    "MissingLeftBracketError": handle_MissingLeftBracketError,
    "MissingRepeatColonError": handle_MissingRepeatColonError,
    "MissingRepeatError": handle_MissingRepeatError,
    "NobreakFirstError": handle_NobreakFirstError,
    "NobreakSyntaxError": handle_NobreakSyntaxError,
    "RepeatFirstError": handle_RepeatFirstError,
    "TabError": handle_TabError,
    "TryNobreakError": handle_TryNobreakError,
    "UnknownLanguageError": handle_UnknownLanguageError,
    "UnknownDialectError": handle_UnknownDialectError,
}



def test_unknown_lang(capsys):
    from . import raise_unknown_language

    if not hasattr(raise_unknown_language, "err"):
        info = str(capsys.readouterr())
        raise_unknown_language.err = info
    assert "UnknownLanguageError" in raise_unknown_language.err



"""pyen.py
"""

pyen = {
    "False": "False",
    "None": "None",
    "True": "True",
    "and": "and",
    "as": "as",
    "assert": "assert",
    "async": "async",  # do not translate
    "await": "await",  # as these are not for beginners
    "break": "break",
    "class": "class",
    "continue": "continue",
    "def": "def",
    "del": "del",
    "elif": "elif",
    "else": ["else", "nobreak"],  # Special case
    "except": "except",
    "finally": "finally",
    "for": "for",
    "from": "from",
    "global": "global",
    "if": "if",
    "import": "import",
    "in": "in",
    "is": "is",
    "lambda": "function",  # Clearer for beginners
    "nonlocal": "nonlocal",
    "not": "not",
    "or": "or",
    "pass": "pass",
    "raise": "raise",
    "return": "return",
    "try": "try",
    "while": "while",
    "with": "with",
    "yield": "yield",
    # Special loop keywords
    "repeat": "repeat",
    "forever": "forever",
    "until": "until",
    # a few builtins useful for beginners
    "input": "input",
    "print": "print",
    "range": "range",
    "exit": "exit",
    # a well-known python expression
    '__name__ == "__main__"': "notimported",
}



def test_missingleftbracket(capsys):
    from . import raise_missing_left_bracket

    if not hasattr(raise_missing_left_bracket, "err"):
        info = str(capsys.readouterr())
        raise_missing_left_bracket.err = info
    assert "MissingLeftBracketError" in raise_missing_left_bracket.err



"""
invocation.py
---------------

AvantPy sets up an import hook which
makes it possible to run a file that contains modified Python syntax,
provided the relevant source transformers can be imported.

Basic invocation
----------------

The simplest invocation of AvantPy sets up the import hook,
collect all existing dialects, and leaves the user in a custom REPL::

    python -m avantpy

This custom REPL should work in a way similar to Python's own.
Please, feel free to file issues for any unexpected behaviour.

If you want to select a particular dialect/language to be available in the
console, you can use the ``--lang`` flag::

    python -m avantpy --lang fr

If a source program is to be run, as described below, the ``--lang`` flag
is ignored.

Running programs
----------------

The primary role of AvantPy is to run programs that have a modified syntax.
This is done by one of the two following alternatives::

    python -m avantpy -s path.to.file
    python -m avantpy --source path.to.file

.. warning::

    Do not include the extension in path.to.file.

You can use Python's interactive flag, either separately,
as in ``python -i -m ...`` or combined with the ``-m`` flag
as done below, to execute a program to be run as "main" and
continue with the console.

The following example is run from the root folder of the AvantPy repository.
The file that is run ends with the ``pyfr`` extension which AvantPy uses
to recognize that the French dialect is to be used::

    $ python -im avantpy -s tests.pyfr.test_french
    Success.
    avantpy console version 0.0.5. [Python version: 3.7.0]

    ->> si Vrai:
    ...     afficher(bonjour)
    ...
    Bonjour tout le monde !
    ->>

"""
import argparse
import sys

from . import session
from . import exception_handling
from . import import_hook
from . import gui
from . import console


parser = argparse.ArgumentParser(
    description="""AvantPy sets up an import hook which
            makes it possible to run a file that contains modified Python syntax
            provided the relevant source transformers can be imported.
        """
)
parser.add_argument(
    "-s",
    "--source",
    help="""Source file to be transformed and executed.
            It is assumed that it can be imported.
            Format: path.to.file -- Do not include an extension.
         """,
)
parser.add_argument(
    "--lang",
    help="""This sets the language used by AvantPy.
            Usually this is a two-letter code such as 'fr' for French.
            If DIALECT is not specified, this will also sets the corresponding
            DIALECT.
         """,
)

parser.add_argument(
    "--dialect",
    help="""This sets the dialect used by AvantPy.
            Usually this is a two-letter code such as 'pyfr' for French.
            If LANG is not specified, this will also sets the corresponding
            value for LANG.
         """,
)

parser.add_argument(
    "--dev_py",
    help="""This disables the custom exception handling so that Python
            tracebacks are printed.
         """,
    action="store_true",
)

parser.add_argument(
    "--show_converted",
    help="""When using the console, if this flag is set, each time the
            code entered is compaeed with the code transformed. If the
            two are not identical, the converted code is printed in
            the console.
         """,
    action="store_true",
)

parser.add_argument(
    "--gui",
    help="""Launches a basic GUI interface, useful for some converting
            programs from one dialect into another or into Python.
         """,
    action="store_true",
)


def main():
    console_dict = {}
    show_python = False
    args = parser.parse_args()

    if args.lang is not None:
        session.state.set_lang(args.lang)

    if args.dialect is not None:
        session.state.set_dialect(args.dialect)

    if args.show_converted:
        show_python = True

    if args.dev_py:
        exception_handling.disable()

    if args.source is not None:
        try:
            main_module = import_hook.import_main(args.source)
            if sys.flags.interactive:
                main_dict = {}
                for var in dir(main_module):
                    if var in ["__cached__", "__loader__", "__package__", "__spec__"]:
                        continue
                    main_dict[var] = getattr(main_module, var)
                console.start_console(local_vars=main_dict, show_python=show_python)
        except ModuleNotFoundError:
            print("Could not find module ", args.source, "\n")
            raise
    elif args.gui:
        gui.main()
    else:
        console.start_console(local_vars=console_dict, show_python=show_python)


if "-m" in sys.argv:
    main()



def test_missing_repeat_colon(capsys):
    from . import raise_missing_repeat_colon

    if not hasattr(raise_missing_repeat_colon, "err"):
        info = str(capsys.readouterr())
        raise_missing_repeat_colon.err = info
    assert "MissingRepeatColonError" in raise_missing_repeat_colon.err



"""my_gettext.py

The usual pattern when using gettext is to surround strings to be translated
by a call to a function named _, as in

    _("This string should be translated.")

This is done when having gettext "install" a given language: it adds a function
named _ to the global builtins. However, this can fail if some other program,
such as the Python REPL, also modifies the builtins and define _ in its own
ways.

To avoid such problems, we use a custom class that keep track of the language
preferred for the translation.  Inside any namespace, when we need to
provide a translation, we define locally _ to be

    _ = gettext_lang.lang

where current_lang.lang means gettext.translation().gettext where
gettext.translation() is the class-based API for gettext.
"""


import gettext
import os


class LangState:
    def __init__(self):
        self.lang = None

    def install(self, lang):
        """Sets the language to be used for translations"""
        if lang is None:
            lang = "en"
        try:
            # We first look for the exact language requested.
            _lang = gettext.translation(
                "avantpy",
                localedir=os.path.normpath(
                    os.path.join(os.path.dirname(__file__), "locales")
                ),
                languages=[lang],
                fallback=False,
            )
        except FileNotFoundError:
            # If it is not available, we make it possible to replace a
            # language specific to a region, as in fr_CA, by a more
            # generic version, such as fr, defined by a two-letter code.
            lang = lang[:2]
            _lang = gettext.translation(
                "avantpy",
                localedir=os.path.normpath(
                    os.path.join(os.path.dirname(__file__), "locales")
                ),
                languages=[lang],
                fallback=True,  # This means that the hard-coded strings in
                # the source file will be used if the requested language
                # is not available.
            )
        self.lang = _lang.gettext


gettext_lang = LangState()  # noqa



"""A custom importer making use of the import hook capability
"""
import importlib
import os
import os.path
import sys

from codeop import CommandCompiler
from importlib.abc import Loader, MetaPathFinder
from importlib.util import spec_from_file_location

import friendly_traceback

from . import session
from . import converter
from .my_gettext import gettext_lang

MAIN_MODULE_NAME = None
COUNTER = 0


def import_main(name):
    """Imports the module that is to be interpreted as the main module.

       avantpy is often invoked with a script meant to be run as the
       main module its source is transformed with the -s (or --source) option,
       as in::

           python -m avantpy -s name

       Python identifies avantpy as the main script, which is not what we want.
    """
    global MAIN_MODULE_NAME
    _ = gettext_lang.lang
    MAIN_MODULE_NAME = name
    try:
        main = importlib.import_module(name)
        return main
    except ModuleNotFoundError:
        print(_("Cannot find main module: "), name)


class AvantPyMetaFinder(MetaPathFinder):
    """A custom finder to locate modules.  The main reason for this code
       is to ensure that our custom loader, which does the code transformations,
       is used."""

    def find_spec(self, fullname, path, target=None):
        """Finds the appropriate properties (spec) of a module, and sets
           its loader."""

        if not path:
            path = [os.getcwd()]
        if "." in fullname:
            name = fullname.split(".")[-1]
        else:
            name = fullname
        for entry in path:
            for ext in session.state.all_dialects():
                filename = os.path.join(entry, name + "." + ext)
                if os.path.exists(filename):
                    break
            else:
                continue

            return spec_from_file_location(
                fullname,
                filename,
                loader=AvantPyLoader(filename),
                submodule_search_locations=None,
            )
        return None  # Not an AvantPy file


sys.meta_path.insert(0, AvantPyMetaFinder())


class AvantPyLoader(Loader):
    """A custom loader which will transform the source prior to its execution"""

    def __init__(self, filename):
        self.filename = filename
        self.compile = CommandCompiler()

    def exec_module(self, module):
        """import the source code, converts it before executing it."""
        global COUNTER
        COUNTER += 1
        if module.__name__ == MAIN_MODULE_NAME:
            module.__name__ = "__main__"

        with open(self.filename, encoding="utf8") as f:
            source = f.read()
        # original = source

        _path, extension = os.path.splitext(self.filename)
        # name = os.path.basename(_path)
        # fullname = name + extension
        dialect = extension[1:]

        try:
            session.state.set_dialect(dialect)
            source = converter.convert(source, dialect, filename=self.filename)
        except Exception:
            friendly_traceback.explain()
            return

        # ------------------------
        # Ideally, instead of the following use of exec(source), we would
        # proceed in two steps:
        # 1. use a custom AST parser that could generate more detailed
        #    information when a SyntaxError is found
        # 2. If no error is found, exec the code objects produced by the AST.
        # -------------------------

        try:
            exec(source, vars(module))
        except Exception:
            # exec() always gives "<string>" as file name
            friendly_traceback.utils.cache_string_source(
                "<string>", (self.filename, source)
            )
            # We remove the first stack item because it is our own code.
            sys.last_type, sys.last_value, last_tb = sys.exc_info()
            sys.last_traceback = last_tb
            try:
                friendly_traceback._explain(
                    sys.last_type, sys.last_value, last_tb.tb_next
                )
            finally:
                last_tb = None
        return



"""replicate_tests.py

Converts all the tests found in tests/pyupper into each dialect
and adds them to the apprpriate folder.

IMPORTANT: Some files cannot be replicated as they raise exceptions.
In these cases, the copies need to be edited "by hand".
"""

import glob
import os
import sys


this_dir = os.path.dirname(__file__)
root_dir = os.path.abspath(os.path.join(this_dir, ".."))
source_dir = os.path.abspath(os.path.join(this_dir, "pyupper"))

sys.path.insert(0, root_dir)

# set up import hook
import avantpy  # noqa

pyupper_files = glob.glob(source_dir + "/*")
dialects = ["pyen", "pyes", "pyfr"]

nb_new = 0
nb_updated = 0

for filename in pyupper_files:
    path, ext = os.path.splitext(filename)
    basename = os.path.basename(path)
    if basename.startswith("__p"):
        continue
    source_creation_time = os.path.getmtime(filename)
    with open(filename, encoding="utf8") as f:
        source = f.read()
    for dialect in dialects:
        error_msg = None
        other_file = filename.replace("pyupper", dialect)
        if ext == ".py":
            content = source
        else:
            try:
                content = avantpy.converter.transcode(source, "pyupper", dialect)
            except Exception as exc:
                error_msg = repr(exc) + "\ncouldn't convert " + filename
                error_msg += "\n   Simply copied the content of the original file."
                content = source
        if os.path.isfile(other_file):
            target_creation_time = os.path.getmtime(other_file)
            newer = source_creation_time > target_creation_time
            if newer:
                nb_updated += 1
                with open(other_file, "w", encoding="utf8") as new_file:
                    new_file.write(content)
        else:
            nb_new += 1
            if error_msg is not None:
                print(error_msg)
                print("   Created: ", other_file)
                print("     IMPORTANT: You will likely need to edit this file.")
            with open(other_file, "w", encoding="utf8") as new_file:
                new_file.write(content)

print("files updated:", nb_updated)
print("new files:", nb_new)
